import 'dart:async';
import 'dart:convert';
import 'dart:html' as html;
import 'dart:ui_web' as ui;

import 'package:flutter/material.dart';
import 'package:flutter_web_plugins/flutter_web_plugins.dart';
import 'meet_hour_platform_interface.dart';
import 'package:js/js.dart';

import 'meet_hour_external_api.dart' as meethour;
import 'room_name_constraint.dart';
import 'room_name_constraint_type.dart';

/// MeetHourPlugin Web version for MeetHour Video conference
class MeetHourPlugin extends MeetHourPlatform {
  // List<MeetHourMeetingListener> _listeners = <MeetHourMeetingListener>[];
  // Map<String, MeetHourMeetingListener> _perMeetingListeners = {};

  /// `MeetHourExternalAPI` holder
  meethour.MeetHourAPI? api;

  /// Flag to indicate if external JS are already added
  /// used for extra scripts
  bool extraJSAdded = false;

  /// Regex to validate URL
  RegExp cleanDomain = RegExp(r"^https?:\/\/");

  MeetHourPlugin._() {
    _setupScripts();
  }

  static final MeetHourPlugin _instance = MeetHourPlugin._();

  /// Registry web plugin
  static void registerWith(Registrar registrar) {
    MeetHourPlatform.instance = _instance;
  }

  /// Joins a meeting based on the MeetHourMeetingOptions passed in.
  /// A MeetHourMeetingListener can be attached to this meeting
  /// that will automatically be removed when the meeting has ended
  @override
  Future<MeetHourMeetingResponse> joinMeeting(MeetHourMeetingOptions options,
      {MeetHourMeetingListener? listener,
      Map<RoomNameConstraintType, RoomNameConstraint>?
          roomNameConstraints}) async {
    // encode `options` Map to Json to avoid error
    // in interoperability conversions
    String webOptions = jsonEncode(options.webOptions);
    String serverURL = options.serverURL ?? "meethour.io";
    serverURL = serverURL.replaceAll(cleanDomain, "");
    api = meethour.MeetHourAPI(serverURL, webOptions);

    // setup listeners
    if (listener != null) {
      api?.on("videoConferenceJoined", allowInterop((dynamic _message) {
        // Mapping object according with meethour external api source code
        Map<String, dynamic> message = {
          "displayName": _message.displayName,
          "roomName": _message.roomName
        };
        listener.onConferenceJoined?.call(message);
      }));
      api?.on("videoConferenceLeft", allowInterop((dynamic _message) {
        // Mapping object according with meethour external api source code
        Map<String, dynamic> message = {"roomName": _message.roomName};
        listener.onConferenceTerminated?.call(message);
      }));
      api?.on("feedbackSubmitted", allowInterop((dynamic message) {
        debugPrint("feedbackSubmitted message: $message");
        listener.onError?.call(message);
      }));

      // NOTE: `onConferenceWillJoin` is not supported or nof found event in web

      // add geeric listener
      _addGenericListeners(listener);

      // force to dispose view when close meeting
      // this is needed to allow create another room in
      // the same view without reload it
      api?.on("readyToClose", allowInterop((dynamic message) {
        api?.dispose();
      }));
    }

    return MeetHourMeetingResponse(isSuccess: true);
  }

  // add generic lister over current session
  _addGenericListeners(MeetHourMeetingListener listener) {
    if (api == null) {
      debugPrint("MeetHour instance not exists event can't be attached");
      return;
    }
    debugPrint("genericListeners ${listener.genericListeners}");
    if (listener.genericListeners != null) {
      listener.genericListeners?.forEach((item) {
        debugPrint("eventName ${item.eventName}");
        api?.on(item.eventName, allowInterop(item.callback));
      });
    }
  }

  @override
  void executeCommand(String command, List<String> args) {
    api?.executeCommand(command, args);
  }

  closeMeeting() {
    debugPrint("Closing the meeting");
    api?.dispose();
    api = null;
  }

  /// Adds a MeetHourMeetingListener that will broadcast conference events
  addListener(MeetHourMeetingListener mhMeetingListener) {
    debugPrint("Adding listeners");
    _addGenericListeners(mhMeetingListener);
  }

  /// Remove MeetHourListener
  /// Remove all list of listeners bassed on event name
  removeListener(MeetHourMeetingListener mhMeetingListener) {
    debugPrint("Removing listeners");
    List<String> listeners = [];
    if (mhMeetingListener.onConferenceJoined != null) {
      listeners.add("videoConferenceJoined");
    }
    ;
    if (mhMeetingListener.onConferenceTerminated != null) {
      listeners.add("videoConferenceLeft");
    }
    ;
    mhMeetingListener.genericListeners
        ?.forEach((element) => listeners.add(element.eventName));
    api?.removeEventListener(listeners);
  }

  /// Removes all MeetHourMeetingListeners
  /// Not used for web
  removeAllListeners() {}

  void initialize() {}

  @override
  Widget buildView(List<String> extraJS) {
    // ignore: undefined_prefixed_name
    ui.platformViewRegistry.registerViewFactory('meet-hour-view',
        (int viewId) {
      final div = html.DivElement()
        ..id = "meet-hour-section"
        ..style.width = '100%'
        ..style.height = '100%';
      return div;
    });
    // add extraJS only once
    // this validation is needed because the view can be
    // rebuileded several times
    if (!extraJSAdded) {
      _setupExtraScripts(extraJS);
      extraJSAdded = true;
    }

    return HtmlElementView(viewType: 'meet-hour-view');
  }

  // setu extra JS Scripts
  void _setupExtraScripts(List<String> extraJS) {
    extraJS.forEach((element) {
      RegExp regExp = RegExp(r"<script[^>]*>(.*?)<\/script[^>]*>");
      if (regExp.hasMatch(element)) {
        final html.NodeValidatorBuilder validator =
            html.NodeValidatorBuilder.common()
              ..allowElement('script',
                  attributes: ['type', 'crossorigin', 'integrity', 'src']);
        debugPrint("ADD script $element");
        html.Element script = html.Element.html(element, validator: validator);
        html.querySelector('head')?.children.add(script);
        // html.querySelector('head').appendHtml(element, validator: validator);
      } else {
        debugPrint("$element is not a valid script");
      }
    });
  }

  // Setup the `MeetHourExternalAPI` JS script
  void _setupScripts() {
    final html.ScriptElement script = html.ScriptElement()
      ..appendText(_clientJs());
    html.querySelector('head')?.children.add(script);
  }

  // Script to allow MeetHour interaction
  // To allow Flutter interact with `MeetHourExternalAPI`
  // extends and override the constructor is needed
  String _clientJs() => """
class MeetHourAPI extends MeetHourExternalAPI {
    constructor(domain , options) {
      console.log(options);
      var _options = JSON.parse(options);
      if (!_options.hasOwnProperty("width")) {
        _options.width='100%';
      }
      if (!_options.hasOwnProperty("height")) {
        _options.height='100%';
      }
      // override parent to atach to view
      //_options.parentNode=document.getElementsByTagName('flt-platform-vw')[0].shadowRoot.getElementById('meet-hour-section');
      console.log(_options);
      _options.parentNode=document.querySelector("#meet-hour-section");
      super(domain, _options);
    }
}
var meethour = { MeetHourAPI: MeetHourAPI };""";
}
